package example3;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Savepoint;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicInteger;

import javax.sql.XAConnection;
import javax.sql.XADataSource;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import com.logntw.eva.orm.DBConnectionProvider;
import com.logntw.eva.sql.dialect.Dialect;

public class XAConnectionProvider implements DBConnectionProvider
{
	private final XADataSource datasource;
	private final Dialect dialect;
	private final int formatId;
	public XAConnectionProvider(int formatId, Dialect dialect, XADataSource datasource)
	{
		if(formatId < 0 || dialect == null || datasource == null)
			throw new java.lang.IllegalArgumentException();
		this.formatId = formatId;
		this.dialect = dialect;
		this.datasource = datasource;
	}
	
	private final class MyXid implements Xid
	{			
		private final byte[] globaltransactionId = new byte[4];
		private final byte[] branchId = new byte[4];
		public MyXid(int gid, int bid)
		{
			this.globaltransactionId[0] = (byte)( (gid & 0xff000000) >>> 24);
			this.globaltransactionId[1] = (byte)( (gid & 0x00ff0000) >>> 16);
			this.globaltransactionId[2] = (byte)( (gid & 0x0000ff00) >>> 8);
			this.globaltransactionId[3] = (byte)( (gid & 0x000000ff) );
			this.branchId[0] = (byte)( (bid & 0xff000000) >>> 24);
			this.branchId[1] = (byte)( (bid & 0x00ff0000) >>> 16);
			this.branchId[2] = (byte)( (bid & 0x0000ff00) >>> 8);
			this.branchId[3] = (byte)( (bid & 0x000000ff) );
		}
		
		@Override
		public int getFormatId()
		{			
			return formatId;
		}
		
		@Override
		public byte[] getBranchQualifier()
		{			
			return branchId;
		}		

		@Override
		public byte[] getGlobalTransactionId()
		{			
			return globaltransactionId;
		}		
		
		private XAConnection connection;
		public XAConnection getConnection()
		{
			return connection;
		}
		public void setConnection(XAConnection connection)
		{
			this.connection = connection;
		}	
	}
	
	private AtomicInteger branchId = new AtomicInteger();
	private MyXid newXid(Integer globalTransactionID)
	{
		MyXid xid = new MyXid(globalTransactionID, branchId.incrementAndGet());
		branchId.compareAndSet(Integer.MAX_VALUE, 0);
		MyXid that = Xids.putIfAbsent(globalTransactionID, xid);
		if(that != null)
			return that;
		return xid;
	}
	
	private final ConcurrentMap<Integer, MyXid> Xids = new ConcurrentHashMap<Integer,MyXid>();	
	@Override
	public Connection getConnection(Object key)throws SQLException
	{		
		Integer globalTransactionID = (Integer)key;
		if(globalTransactionID == null || globalTransactionID < 0)
			throw new IllegalArgumentException();
		MyXid xid = Xids.get(globalTransactionID);
		if(xid == null)
			xid = newXid(globalTransactionID);		
		XAConnection xaconnection = datasource.getXAConnection();
		Connection connection = xaconnection.getConnection();
		xid.setConnection(xaconnection);
		
		XAResource xaRes = xaconnection.getXAResource();		
		try
		{
			xaRes.start(xid, XAResource.TMNOFLAGS);
		}
		catch (XAException e)
		{
			throw new SQLException(e.getMessage(), e);
		} 

		return connection;
	}
		
	@Override
	public boolean commit(Object key, Connection connection) throws SQLException
	{
		Integer globalTransactionID = (Integer)key;
		MyXid xid = Xids.get(globalTransactionID);
		if(xid == null)
			throw new java.lang.IllegalStateException();
		XAConnection xaconnection = xid.getConnection();
		
		XAResource xaRes = xaconnection.getXAResource();		
		try
		{
			xaRes.end(xid, XAResource.TMSUCCESS);
			int ret = xaRes.prepare(xid); 
			if (ret == XAResource.XA_OK)  			
				xaRes.commit(xid, false); 			
		}
		catch (XAException e)
		{
			throw new SQLException(e.getMessage(), e);
		} 		
		return false;
	}

	@Override
	public boolean rollback(Object key, Connection connection) throws SQLException
	{
		Integer globalTransactionID = (Integer)key;
		MyXid xid = Xids.get(globalTransactionID);
		if(xid == null)
			throw new java.lang.IllegalStateException();
		XAConnection xaconnection = xid.getConnection();
		
		XAResource xaRes = xaconnection.getXAResource();		
		try
		{
			xaRes.end(xid, XAResource.TMSUCCESS);
			int ret = xaRes.prepare(xid); 
			if (ret == XAResource.XA_OK)			
				xaRes.rollback(xid);
		}
		catch (XAException e)
		{
			throw new SQLException(e.getMessage(), e);
		} 		
		return false;
	}

	@Override
	public boolean rollback(Object key, Connection connection, Savepoint savepoint) throws SQLException
	{
		String msg = "cannot set SAVEPOINT in a distributed transaction";
		throw new java.lang.IllegalStateException(msg);
	}

	@Override
	public boolean releaseConnection(Object key, Connection connection) throws SQLException
	{		
		Xids.remove(key);
		return true;
	}	

	@Override
	public boolean commit(Connection connection) throws SQLException
	{		
		return false;
	}

	@Override
	public Dialect getDialect()
	{
		return dialect;
	}
	
	@Override
	public void dispose()
	{
		// do nothing.
	}

	@Override
	public Connection getConnection() throws SQLException
	{
		return datasource.getXAConnection().getConnection();
	}

	@Override
	public boolean releaseConnection(Connection connection) throws SQLException
	{
		return true;
	}

	@Override
	public boolean rollback(Connection connection) throws SQLException
	{
		return true;
	}

	@Override
	public boolean rollback(Connection connection, Savepoint savepoint) throws SQLException
	{		
		return true;
	}
}